/*
 *  (c) 2008 Andreas Kiefer, disco | Distributed Computer Systems Lab
 *                                   University of Kaiserslautern, Germany
 *      All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import unikl.disco.misc.Rational;
import unikl.disco.misc.SetUtils;

import unikl.disco.curves.Curve;

import unikl.disco.nc.Configuration.Flag;
import unikl.disco.network.Link;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Server;

/**
 * @author Frank A. Zdarsky
 * @author Andreas Kiefer
 */
public class PmooSfaAnalysis extends Analysis
{
	private Curve beta_e2e = Curve.createZeroDelayBurst();
	
	@SuppressWarnings("unused")
	private PmooSfaAnalysis(){}
	
	public PmooSfaAnalysis( Network network ) {
		super.network = network;
	}

	/** The current recursion level */
	protected int		rec_depth;

	/**
	 * Performs a pay-multiplexing-only-once (PMOO) analysis for the
	 * <code>flow_of_interest</code> and returns the end-to-end service curve.
	 * 
	 * @param flow_of_interest
	 *            the flow for which the end-to-end service curve shall be
	 *            computed
	 * @return the PMOO end-to-end service curve
	 */
	@Override
	public void performEnd2EndAnalysis( Flow flow_of_interest ) throws Exception
	{
		beta_e2e = Curve.createZeroDelayBurst();
	
		try{
			super.flow_of_interest = flow_of_interest;
			this.rec_depth = -1;
			List<Link> path = flow_of_interest.getPathInclExplSnk();
			beta_e2e = getPMOOServiceCurve( path, new HashSet<Flow>( Collections.singleton( flow_of_interest ) ) );
		} catch (Exception e){
			throw e;
		}
	}

	/**
	 * Returns the PMOO service curve for a set of
	 * <code>flows_of_interest</code>. This service curve is a concatenation of
	 * all servers on the given link path <code>path</code> that have outgoing
	 * links contained in the path (i.e. the server at which the path ends is
	 * not included).<br/>
	 * For performance reasons, the method first computes various mappings of
	 * servers to incoming and leaving flows. It then computes the output bounds
	 * of all incoming flows. If there are rejoining flows (flows that were
	 * there before, then left and now join again), these are substituted for
	 * fresh flows. Finally, <code>computePMOOServiceCurve()</code> is invoked,
	 * which does the real work of concatenating the service curves.<br/>
	 * The output bounds are computed using either the standard
	 * <code>computeOutputBound()</code> method that is also used for the total
	 * flow and separated flow analyses or the new
	 * <code>computePMOOOutputBound()</code> method that results in tighter
	 * bounds but also increases computation effort. The latter recursively
	 * calls this method again. It is also possible to use a mixture of both,
	 * using the PMOO output bound for the first levels of recursion and the
	 * standard output bound for deeper levels. The recursion level from which
	 * on the standard bound is used may be set via
	 * <code>setMaxRecursionDepthForPMOO()</code>. The default is to always use
	 * PMOO. To always use the standard output bound computation set the maximum
	 * recursion depth to 0.<br/>
	 * Several techniques that significantly reduce the complexity of the
	 * service curve computation are employed:
	 * <ul>
	 * <li><it>Flow Aggregation</it>: Flows having the same ingress and egress
	 * points are always replaced by a single flow whose bound is the aggregate
	 * bound of all contained flows.</li>
	 * <li><it>Arrival Curve Approximation<it>: Approximates the arrival curves
	 * of incoming flows by a different arrival curve with a lower number of
	 * linear segments. This does have an effect on the output bound, but does
	 * not necessarily affect the delay bound. To enable this technique, use
	 * <code>setUseArrivalCurveApproximation()</code>.</li>
	 * </ul>
	 * 
	 * @param path
	 *            the link path along which service curves are concatenated (not
	 *            including the server at which the path ends)
	 * @param flows_of_interest
	 *            the set of flows of interest
	 * @return the concatenated PMOO service curve
	 * @throws Exception 
	 * @see NetworkAnalyser#setMaxRecursionDepthForPMOO(int)
	 * @see NetworkAnalyser#setUseArrivalCurveApproximation(boolean)
	 */
	protected Curve getPMOOServiceCurve( List<Link> path, Set<Flow> flows_of_interest ) throws Exception
	{
		rec_depth++;

		List<Set<Flow>> joining_flow_sets = network.getJoiningFlowLists( path, flows_of_interest );
		List<Set<Flow>> rejoining_flow_list = network.getRejoiningFlowLists( joining_flow_sets );
		Set<Flow> all_interfering_flows = SetUtils.getUnion( joining_flow_sets );
		Map<Flow, Integer> flow_egress_map = network.getFlowEgressMap( path, all_interfering_flows );
		Map<Flow, Curve> flow_bound_map = new HashMap<Flow, Curve>();

		int i = -1;
		for ( Link l : path )
		{
			i++;
			Server server = l.getSource();
			
			if( Configuration.getUseArbMuxFlag() == Flag.GLOBALLY_OFF
				|| ( Configuration.getUseArbMuxFlag() == Flag.SERVERLOCAL_FLAG & server.useFifoMultiplexing() ) )
			{
				throw new Exception();
			}

			Set<Flow> joining_flows = joining_flow_sets.get( i );
			Set<Flow> rejoining_flows = rejoining_flow_list.get( i );

			Set<Set<Flow>> aggregateable_flow_sets = network.getAggregateableFlowSets(	path,
																					i,
																					joining_flows,
																					flow_egress_map );
			for ( Set<Flow> aggregateable_flows : aggregateable_flow_sets )
			{
				// search for subsets of the aggregate on all incoming links;
				// computePBOOOutputBound()
				// will correctly handle the case of empty subsets on an
				// incoming link
				Curve aggregate_bound = Curve.createNullCurve();
				Set<Flow> inlink_flows = network.getSourceFlowSet( server );
				Set<Flow> flow_set = SetUtils.getIntersection( inlink_flows, aggregateable_flows );
				
				aggregate_bound = Curve.add( aggregate_bound, network.getAggregateSourceFlowAC( server, flow_set ) );
			
				for ( Link in_l : network.getInLinks( server) )
				{
					inlink_flows = network.getLinkFlowSet( in_l );
					flow_set = SetUtils.getIntersection( inlink_flows, aggregateable_flows );
					
					Curve alpha_excl;
					if ( rec_depth < Configuration.getMaxRecursionDepthForPMOO() )
					{
						alpha_excl = computePMOOOutputBound( server, flow_set );
					}
					else
					{
						alpha_excl = computeOutputBound( in_l.getSource(), flow_set );
					}
					aggregate_bound = Curve.add( aggregate_bound, alpha_excl );
				}

				Set<Flow> rejoining_flows_in_aggregate = SetUtils.getIntersection(aggregateable_flows, rejoining_flows );
				if ( rejoining_flows_in_aggregate.size() > 0 )
				{
					int egress_server = network.findEgressPoint( path,
															i,
															rejoining_flows_in_aggregate.iterator().next() );
					network.createSubstituteFlows(	rejoining_flows_in_aggregate,
												egress_server,
												aggregateable_flows,
												all_interfering_flows,
												joining_flows,
												flow_egress_map );
				}

				Flow aggregate_flow;
				if ( aggregateable_flows.size() == 1 )
				{
					// store bounds for unaggregateable flows directly
					aggregate_flow = aggregateable_flows.iterator().next();
				}
				else
				{
					aggregate_flow = network.aggregateFlows(	aggregateable_flows,
														joining_flows,
														all_interfering_flows,
														flow_egress_map );
				}
				flow_bound_map.put( aggregate_flow, aggregate_bound );
			}
		}

		rec_depth--;
		return computePMOOServiceCurve( path,
										all_interfering_flows,
										joining_flow_sets,
										flow_egress_map,
										flow_bound_map,
										rec_depth + 1 );
	}

	/**
	 * Computes the PMOO output bound for a set of
	 * <code>flows_of_interest</code>. The difference to the standard output
	 * bound method is that this method tries to compute tighter bounds by
	 * concatenating as many servers as possible using the PMOO approach. It
	 * does so by searching from <code>server</code> towards the sinks of the
	 * flows contained in <code>flows_of_interest</code> until it reaches the
	 * server where all these flows first meet each other (the
	 * "splitting point"). It then concatenates all servers between the
	 * splitting point (inclusive) and <code>server</code> (exclusive).<br/>
	 * Note: The previous sentence means that the output bound is for the flows
	 * <b>before</b> <code>server</code> and not after the server as is the case
	 * with the standard computateOutputBound().
	 * 
	 * @param server
	 *            that all flows of interest flow into
	 * @param flows_of_interest
	 *            the set of flows of interest
	 * @return the PMOO output bound
	 * @throws Exception 
	 */
	protected Curve computePMOOOutputBound( Server server, Set<Flow> flows_of_interest ) throws Exception
	{
		Curve alpha_out = Curve.createNullCurve();

		if ( flows_of_interest.size() == 0 )
		{
			return alpha_out;
		}
		
		Server split = network.findSplittingPoint( server, flows_of_interest );
		if ( split.equals( server ) )
		{
			throw new RuntimeException( "could not find splitting point..." );
		}

		Flow f = flows_of_interest.iterator().next();
		if ( split.equals( f.getSource() ) )
		{
			Curve alpha = network.getAggregateSourceFlowAC( f.getSource(), flows_of_interest );

			List<Link> path = f.getSubPath( f.getSource(), server);

			if ( path.isEmpty() )
			{
				alpha_out = alpha;
			}
			else
			{
				Curve beta;
				try {
					beta = getPMOOServiceCurve( path, flows_of_interest );
				} catch (Exception e) {
					throw e;
				}

				if( Configuration.getUseGammaFlag() == Flag.SERVERLOCAL_FLAG )
				{
					Curve gamma = getGammaOnPath( path );
					alpha_out = Curve.makeArrivalCurve( MinPlus.deconvolve( MinPlus.convolve( alpha, gamma ), beta ) );
				}

				if( Configuration.getUseGammaFlag() == Flag.GLOBALLY_ON )
				{
					Curve gamma = getMaxServiceCurveOnPath( path );
					alpha_out = Curve.makeArrivalCurve( MinPlus.deconvolve( MinPlus.convolve( alpha, gamma ), beta ) );
				}
				
				if( Configuration.getUseExtraGammaFlag() == Flag.SERVERLOCAL_FLAG )
				{
					Curve extra_gamma = getExtraGammaOnPath( path );
					alpha_out = MinPlus.convolve( alpha_out, extra_gamma );
				}

				if( Configuration.getUseExtraGammaFlag() == Flag.GLOBALLY_ON )
				{
					Curve extra_gamma = getMaxServiceCurveOnPath( path );
					alpha_out = MinPlus.convolve( alpha_out, extra_gamma );
				}
			}
		}
		else
		{
			List<Link> path = f.getSubPath( split, server );

			Curve alpha = Curve.createNullCurve();
			for ( Link in_l : network.getInLinks(split) )
			{
				Set<Flow> inflows = SetUtils.getIntersection( network.getLinkFlowSet( in_l ), flows_of_interest );
				alpha = Curve.add( alpha, computePMOOOutputBound( split, inflows ) );
			}
			Curve beta = getPMOOServiceCurve( path, flows_of_interest );

			Curve gamma = getGammaOnPath( path );
			alpha_out = Curve.makeArrivalCurve( MinPlus.deconvolve( MinPlus.convolve( alpha, gamma ), beta ) );
			
			Curve extra_gamma = getExtraGammaOnPath( path );
			alpha_out = MinPlus.convolve( alpha_out, extra_gamma );
		}
		return alpha_out;
	}

	/**
	 * Concatenates the service curves along the given path <code>path</code>
	 * according to the PMOO approach and returns the result.<br/>
	 * It first decomposes all arrival curves (service curves) into token
	 * buckets (rate latency curves), enumerates over all combinations of token
	 * buckets and rate latency curves, and calls
	 * <code>computePartialPMOOServiceCurve()</code> for each combination. The
	 * total PMOO service curve is the maximum of all partial service curves.
	 * 
	 * @param path
	 *            link path traversing the servers whose service curves shall be
	 *            concatenated (excluding the one at which the path ends).
	 * @param all_interfering_flows
	 *            all flows that cross one of the servers on the path and do not
	 *            belong to the flows of interest
	 * @param joining_flow_sets
	 *            a list of sets with the flows that join at each server
	 * @param flow_egress_map
	 *            map of flows to egress points (integer index into the path)
	 * @param flow_bound_map
	 *            map of flows to their arrival curves
	 * @param rec_depth
	 *            the current recursion level (optional)
	 * @return the PMOO service curve
	 */
	protected Curve computePMOOServiceCurve(	List<Link> path,
												Set<Flow> all_interfering_flows,
												List<Set<Flow>> joining_flow_sets,
												Map<Flow, Integer> flow_egress_map,
												Map<Flow, Curve> flow_bound_map,
												int rec_depth )
	{
		
		// create a flow-->tb_iter map
		Map<Flow, Integer> flow_tb_iter_map = new HashMap<Flow, Integer>();
		for ( Flow f : all_interfering_flows )
		{
			flow_tb_iter_map.put( f, Integer.valueOf( 0 ) );
		}
		// create a list of rl_iters
		Curve[] service_curves = new Curve[path.size()];
		int[] server_rl_iters = new int[path.size()];
		int[] server_rl_counts = new int[path.size()];
		int i = 0;
		for ( Link l : path )
		{
			Server server = l.getSource();
			Curve service_curve = server.getServiceCurve();
			service_curves[i] = service_curve;
			server_rl_iters[i] = 0;
			server_rl_counts[i] = service_curve.getRLComponentCount();
			i++;
		}

		Curve beta_total = Curve.createNullCurve();

		boolean more_combinations = true;
		while ( more_combinations )
		{
			// compute service curve for this combination
			Curve beta = computePartialPMOOServiceCurve(	service_curves,
															all_interfering_flows,
															joining_flow_sets,
															flow_egress_map,
															flow_bound_map,
															flow_tb_iter_map,
															server_rl_iters );
			if ( beta != Curve.createNullCurve() )
			{
				beta_total = Curve.max( beta_total, beta );
			}

			// first check whether there are more combinations of flow TBs
			more_combinations = false;
			for ( Entry<Flow, Integer> entry : flow_tb_iter_map.entrySet() )
			{
				Curve f_bound = flow_bound_map.get( entry.getKey() );

				i = flow_tb_iter_map.get( entry.getKey() ).intValue();
				if ( i + 1 < f_bound.getTBComponentCount() )
				{
					flow_tb_iter_map.put( entry.getKey(), Integer.valueOf( i + 1 ) );
					more_combinations = true;
					break;
				}
				else
				{
					flow_tb_iter_map.put( entry.getKey(), Integer.valueOf( 0 ) );
				}
			}

			// if not, check whether there are more combinations of server RLs
			if ( !more_combinations )
			{
				for ( i = 0; i < server_rl_iters.length; i++ )
				{
					int j = server_rl_iters[i];
					if ( j + 1 < server_rl_counts[i] )
					{
						server_rl_iters[i] = j;
						more_combinations = true;
						break;
					}
					else
					{
						server_rl_iters[i] = 0;
					}
				}
			}
		}

		return beta_total;
	}

	/**
	 * Calculates the partial PMOO service curve for the given flow set by
	 * combining all servers having an outgoing link contained in the given
	 * link-path. For each flow considers only one of its token bucket
	 * components (selected via the flow_tb_iter_map) and for each service curve
	 * considers only one rate latency curve (selected via the server_rl_iters).
	 * 
	 * @return a partial PMOO service curve
	 */
	protected Curve computePartialPMOOServiceCurve( Curve[] service_curves,
													Set<Flow> all_interfering_flows,
													List<Set<Flow>> joining_flow_sets,
													Map<Flow, Integer> flow_egress_map,
													Map<Flow, Curve> flow_bound_map,
													Map<Flow, Integer> flow_tb_iter_map,
													int[] server_rl_iters )
	{

		Rational T = new Rational( 0.0 );
		Rational R = Rational.getPosInfinity();
		Rational sum_bursts = new Rational( 0.0 );
		Rational sum_latencyterms = new Rational( 0.0 );

		Set<Flow> present_flows = new HashSet<Flow>();
		for ( int i = 0; i < service_curves.length; i++ )
		{
			// add incoming flows
			present_flows.addAll( joining_flow_sets.get( i ) );

			Curve current_rl = service_curves[i].getRLComponent( server_rl_iters[i] );

			// sum up latencies
			T = Rational.add( T, current_rl.getLatency() );

			// calculate and store sum of rates of all passing flows
			Rational sum_r = new Rational( 0.0 );
			for ( Flow f : present_flows )
			{
				Curve bound = (Curve) flow_bound_map.get( f );
				Curve current_tb = bound.getTBComponent( ((Integer) flow_tb_iter_map.get( f )).intValue() );
				sum_r = Rational.add( sum_r, current_tb.getSustainedRate() );
			}

			// update latency terms
			sum_latencyterms = Rational.add( sum_latencyterms, Rational.mult( sum_r, current_rl.getLatency() ) );

			// compute residual rate; update min
			Rational Ri = Rational.sub( current_rl.getSustainedRate(), sum_r );
			if ( Ri.le( new Rational( 0.0 ) ) )
			{
				return Curve.createNullCurve();
			}
			R = Rational.min( R, Ri );

			// remove all outgoing flows from the set of present flows
			for ( Iterator<Flow> flow_iter = present_flows.iterator(); flow_iter.hasNext(); )
			{
				Flow f = flow_iter.next();

				int egress_server = ((Integer) flow_egress_map.get( f )).intValue();
				if ( egress_server <= i )
				{
					flow_iter.remove();
				}
			}
		}

		// compute sum of bursts
		for ( Flow f : all_interfering_flows )
		{
			Curve bound = (Curve) flow_bound_map.get( f );
			Curve current_tb = bound.getTBComponent( ((Integer) flow_tb_iter_map.get( f )).intValue() );
			sum_bursts = Rational.add( sum_bursts, current_tb.getTBBurst() );
		}

		T = Rational.add( T,Rational.div( Rational.add( sum_bursts, sum_latencyterms ), R ) );

		return Curve.createRateLatency( R, T );
	}

	/**
	 * Returns the convolution of the maximum service curves of all servers on
	 * the given lnik path <code>path</code> that have the useGamma flag set.<br />
	 * If a server either has no maximum service curve set or useGamma is disabled,
	 * calculations take place with the default maximum service curve,
	 * i.e., the zero delay burst curve, so the result will not be influenced.
	 * 
	 * @param path
	 *            an link path
	 * @return the convolved curve
	 */
	public Curve getGammaOnPath( List<Link> path )
	{
		Curve gamma_total = path.get( 0 ).getSource().getGamma();
		for ( Link l : path )
		{
			Server server = l.getDest();
			gamma_total = MinPlus.convolve( gamma_total, server.getGamma() );
		}
		
		return gamma_total;
	}

	/**
	 * Returns the convolution of the maximum service curves of all servers on
	 * the given link path <code>path</code> that have the useExtraGamma flag set.<br />
	 * If a server either has no maximum service curve set or useExtraGamma is disabled,
	 * calculations take place with the default maximum service curve,
	 * i.e., the zero delay burst curve, so the result will not be influenced.
	 * 
	 * @param path
	 *            an link path
	 * @return the convolved curve
	 */
	public Curve getExtraGammaOnPath( List<Link> path )
	{
		Curve gamma_total = path.get( 0 ).getSource().getExtraGamma();
		for ( Link l : path )
		{
			Server server = l.getDest();
			gamma_total = MinPlus.convolve( gamma_total, server.getExtraGamma() );
		}		
		gamma_total.removeLatency();
		
		return gamma_total;
	}
	
	/**
	 * Returns the convolution of the maximum service curves of all servers on
	 * the given link path <code>path</code><br />
	 * If a server has no maximum service curve, calculations take place with 
	 * the default maximum service curve, i.e., the zero delay burst curve, 
	 * so the result will not be influenced.
	 * 
	 * @param path
	 *            an link path
	 * @return the convolved curve
	 */
	public Curve getMaxServiceCurveOnPath( List<Link> path )
	{
		Curve gamma_total = path.get( 0 ).getSource().getMaxServiceCurve();
		for ( Link l : path )
		{
			Server server = l.getDest();
			gamma_total = MinPlus.convolve( gamma_total, server.getMaxServiceCurve() );
		}
		
		return gamma_total;
	}

	public Curve getEnd2EndServiceCurve () {
		return beta_e2e;
	}
	
	@Override
	public Rational getBacklogBound() {
		return Rational.max( new Rational( 0.0 ),
						Curve.getVerticalDeviation( flow_of_interest.getArrivalCurve(), beta_e2e ));
	}

	@Override
	public Rational getDelayBound() {
		return Rational.max( new Rational( 0.0 ), 
						Curve.getHorizontalDeviation( flow_of_interest.getArrivalCurve(), beta_e2e ));
	}
}
