/*
 *  (c) 2008 Andreas Kiefer, disco | Distributed Computer Systems Lab
 *                                   University of Kaiserslautern, Germany
 *      All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import unikl.disco.misc.Rational;
import unikl.disco.misc.SetUtils;

import unikl.disco.curves.Curve;

import unikl.disco.nc.Configuration.Flag;
import unikl.disco.network.Link;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Server;

/**
 * @author Frank A. Zdarsky
 * @author Andreas Kiefer
 */
public class SeparateFlowAnalysis extends Analysis
{
	Map<Server, Curve> map__server__beta_lofoi = new LinkedHashMap<Server, Curve>();
	Map<Server, Curve> map__server__alpha_xfoi = new LinkedHashMap<Server, Curve>();
	private Curve beta_e2e = Curve.createZeroDelayBurst();

	@SuppressWarnings("unused")
	private SeparateFlowAnalysis(){}
	
	public SeparateFlowAnalysis( Network network ) {
		super.network = network;
	}
	/**
	 * Performs a separated flow analysis for the <code>flow_of_interest</code>
	 * and returns the end-to-end service curve.<br/>
	 * This analysis first blends out the flow of interest and then computes for
	 * each server along this flow's path the effective service curve that
	 * results if all remaining flows crossing this server receive their maximum
	 * amount of service. Then all effective service curves are concatenated to
	 * receive the end-to-end service curve from the perspective of the flow of
	 * interest.
	 * 
	 * @param flow_of_interest
	 *            the flow for which the end-to-end service curve shall be
	 *            computed
	 * @return the end-to-end service curve
	 */
	@Override
	public void performEnd2EndAnalysis( Flow flow_of_interest ) throws Exception
	{
		map__server__beta_lofoi = new LinkedHashMap<Server, Curve>();
		map__server__alpha_xfoi = new LinkedHashMap<Server, Curve>();
		beta_e2e = Curve.createZeroDelayBurst();
		
		try{
			super.flow_of_interest = flow_of_interest;
	
			Set<Flow> flow_of_interest_as_set = new HashSet<Flow>( Collections.singleton( flow_of_interest ) );
	
			List<Link> path = flow_of_interest.getPathInclExplSnk();
			for ( Link l : path )
			{
				Server server = l.getSource();
			
				// First check traffic originating in server
				Set<Flow> f_server = network.getSourceFlowSet( server );				
				Set<Flow> f_xfoi_server = SetUtils.getDifference( f_server, flow_of_interest_as_set );
				Curve alpha_xfoi = network.getAggregateSourceFlowAC( server, f_xfoi_server );
				
				// Second check each incoming links for traffic as well as it's arrival curve seen at server
				for ( Server predecessor : network.getPredecessors( server ) )
				{
					f_server = network.getIncomingFlows( predecessor, server );			
					f_xfoi_server = SetUtils.getDifference( f_server, flow_of_interest_as_set );
					
					if ( predecessor.equals( flow_of_interest.getPredecessingLink( l ).getSource() ) ) {
						alpha_xfoi = Curve.add( alpha_xfoi, computeSfaOutputBound( predecessor, f_xfoi_server, path.indexOf( l )-1 ) );			
					} else {
						alpha_xfoi = Curve.add( alpha_xfoi, computeOutputBound( predecessor, f_xfoi_server ) );						
					}
				}
				
				// Calculate the left-over service curve for the flow of interest
				Curve beta_lofoi = server.getServiceCurve();
				if( Configuration.getUseArbMuxFlag() == Flag.GLOBALLY_OFF
					|| ( Configuration.getUseArbMuxFlag() == Flag.SERVERLOCAL_FLAG & server.useFifoMultiplexing() ) )
				{
					beta_lofoi = computeFifoServiceCurve( alpha_xfoi, beta_lofoi );
				}
				else
				{
					beta_lofoi = Curve.boundAtXAxis( Curve.sub( beta_lofoi, alpha_xfoi ) );				
				}

				map__server__beta_lofoi.put( server, beta_lofoi );
				map__server__alpha_xfoi.put( server, alpha_xfoi );
				
				// Convolve all left over service curves, server by server
				beta_e2e = MinPlus.convolve( beta_e2e, beta_lofoi );
			}
		} catch (Exception e) {
			throw e;		
		}
	}
	
	// Get the aggregated arrival curve of flows on link from server to the node calling this function.
	// the calling node is implicitly given by the outgoing_flows set that resembles the cross flows needed at the caller
	// Note that the caller usually hands over all of it's cross traffic, not only the flows coming from this server
	protected Curve computeSfaOutputBound( Server server, Set<Flow> f_xfcaller, int linkNo_on_path_of_interest ) {
		if ( f_xfcaller.size() == 0 )
		{
			return Curve.createNullCurve();
		}
		
		Curve alpha_xfcaller = Curve.createNullCurve();
		
		// Get flows originating in this server and sort by 
		// flows that are cross traffic at the next server (f_xfcaller_server, i.e, cross traffic of the caller)
		// flow that are cross traffic to them (f_xxfcaller_server, i.e., cross traffic to the cross traffic of the caller)
		Set<Flow> f_server = network.getSourceFlowSet( server );
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
		Set<Flow> f_xxfcaller_server = SetUtils.getDifference( f_server, f_xfcaller_server );
		f_xxfcaller_server.remove( flow_of_interest ); //SFA specific

		Curve alpha_xfcaller_server = network.getAggregateSourceFlowAC( server, f_xfcaller_server );
		Curve alpha_xxfcaller_server = network.getAggregateSourceFlowAC( server, f_xxfcaller_server );

		for ( Server predecessor : network.getPredecessors( server ) )
		{
			f_server = network.getIncomingFlows( predecessor, server );
			f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
			f_xxfcaller_server = SetUtils.getDifference( f_server, f_xfcaller_server );
			f_xxfcaller_server.remove( flow_of_interest );

			// TODO
			// Only direct predecessors are considered but
			// what if the cross flow is crossing the flow of interest's path somewhere earlier again?
			// At the moment this would mean that computeOutputBound was triggered and thus the flow of interest interferes with it's own cross flow
			// See Tandem_1SC_2Flows_1AC_2Paths_v2.java
			if ( linkNo_on_path_of_interest > 0
				 && network.findLink( predecessor, server ) == flow_of_interest.getPath().get( linkNo_on_path_of_interest - 1 ) ) {
				alpha_xfcaller_server = Curve.add( alpha_xfcaller_server, computeSfaOutputBound( predecessor, f_xfcaller_server, linkNo_on_path_of_interest - 1 ) );
				alpha_xxfcaller_server = Curve.add( alpha_xxfcaller_server, computeSfaOutputBound( predecessor, f_xxfcaller_server, linkNo_on_path_of_interest - 1 ) );		
			} else {
				alpha_xfcaller_server = Curve.add( alpha_xfcaller_server, computeOutputBound( predecessor, f_xfcaller_server ) );
				alpha_xxfcaller_server = Curve.add( alpha_xxfcaller_server, computeOutputBound( predecessor, f_xxfcaller_server ) );						
			}
		}

		Curve beta_loxfcaller_server = server.getServiceCurve();
		Curve gamma = server.getMaxServiceCurve();
		if( Configuration.getUseArbMuxFlag() == Flag.GLOBALLY_ON
			|| ( Configuration.getUseArbMuxFlag() == Flag.SERVERLOCAL_FLAG & server.useArbitraryMultiplexing() ) )
		{
			beta_loxfcaller_server = Curve.boundAtXAxis( Curve.sub( beta_loxfcaller_server, alpha_xxfcaller_server ) );
		}
		else
		{
			beta_loxfcaller_server = computeFifoServiceCurve( alpha_xxfcaller_server, beta_loxfcaller_server );
		}

		if( Configuration.getUseGammaFlag() == Flag.GLOBALLY_ON
			|| ( Configuration.getUseGammaFlag() == Flag.SERVERLOCAL_FLAG & server.useGamma() ) )
		{
			alpha_xfcaller = Curve.makeArrivalCurve( MinPlus.deconvolve( MinPlus.convolve( alpha_xfcaller_server, gamma ), beta_loxfcaller_server ) );
		}
		else
		{
			alpha_xfcaller = Curve.makeArrivalCurve( MinPlus.deconvolve( alpha_xfcaller_server, beta_loxfcaller_server ) );
		}
		
		if( Configuration.getUseExtraGammaFlag() == Flag.GLOBALLY_ON
			|| ( Configuration.getUseExtraGammaFlag() == Flag.SERVERLOCAL_FLAG & server.useExtraGamma() ) )
		{
			alpha_xfcaller = MinPlus.convolve( alpha_xfcaller, gamma.removeLatency() );
		}
		
		return alpha_xfcaller;
	}

	public Curve getEnd2EndServiceCurve () {
		return beta_e2e;
	}
	
	@Override
	public Rational getBacklogBound() {
		return Rational.max( new Rational( 0.0 ),
						Curve.getVerticalDeviation( flow_of_interest.getArrivalCurve(), beta_e2e ));
	}
	
	@Override
	public Rational getDelayBound() {
		return Rational.max( new Rational( 0.0 ), 
						Curve.getHorizontalDeviation( flow_of_interest.getArrivalCurve(), beta_e2e ));
	}

	public Map<Server, Curve> getServerLeftOverBetaMap(){
		return map__server__beta_lofoi;
	}

	public Map<Server, Curve> getServerCrossTrafficAlphaMap(){
		return map__server__alpha_xfoi;
	}
}
