/*
 *  (c) 2005,2006 Frank A. Zdarsky, disco | Distributed Computer Systems Lab
 *                                  University of Kaiserslautern, Germany
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.network;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import unikl.disco.curves.Curve;

/**
 * Class representing flows through the network.
 *
 * @author Frank A. Zdarsky
 * @author Steffen Bondorf
 */
public class Flow {
	private static int nextId = 0;

	/** The flow's ID. */
	private Integer	id;
	/** The flow's arrival curve */
	private Curve    ac;
	/** The server at which the flow originates */
	private Server	source;
	/** The server at which the flow ends */
	private Server	sink;
	
	private String alias;
	/** The link path the flow traverses */
	private List<Link> path;

	/**
	 * Creates a dummy flow.
	 * @return a dummy flow
	 */
	public static Flow createDummyFlow() {
		return new Flow();
	}

	/**
	 * @param source the server at which the flow originates
	 * @param sink the server at which the flow ends
	 * @param ac the flow's arrival curve
	 * @param path the link path the flow traverses
	 */
	protected Flow( Curve ac, List<Link> path ) {
		// TODO check if this flow makes sense
		// i.e., links in the path connect source and sink
		// can be done by sorting: If sorting the path fails, it is not connected
		this.id     = Integer.valueOf(nextId++);
		this.ac     = ac;
		
		this.source = path.get( 0 ).getSource();
		this.sink = path.get( path.size()-1 ).getDest();
		
		this.path   = path;
	}
	
	private Flow() {
		this.id     = Integer.valueOf(nextId++);
		this.ac     = null;
		this.source = null;
		this.sink   = null;
		this.path   = null;
	}
	
	public boolean setArrivalCurve( Curve arrival_curve ) {
		ac = arrival_curve;
		return true;
	}

	public Curve getArrivalCurve() {
		return ac;
	}

	public Server getSource() {
		return source;
	}
	
	public Server getSink() {
		return sink;
	}
	
	public String getAlias() {
		return alias;
	}
	
	public void setAlias(String name) {
		alias = name;
	}

	public List<Link> getPathInclExplSnk() {
		return path;
	}

	public List<Link> getPath() {
		return path.subList( 0, path.size()-1 );
	}
	
	public Link getPredecessingLink( Link e ) {
		if( path.indexOf( e ) > 0 ) {
			return path.get( path.indexOf( e ) - 1 );
		} else {
			return new Link( new Server(), new Server() );
		}		
	}
	
    /**
     * Returns a sublist view of the given link path <code>path</code> from the server
     * <code>from</code> to server <code>to</code> or an empty list if the servers are not
     * both on the path and traversed in order.
     * @param path
     * @param from
     * @param to
     * @return the sub path from <code>from</code> to <code>to</code>
     */
    public List<Link> getSubPath(Server from, Server to) {
    	if( from == to ) {
    		return Collections.emptyList();
    	}
    	
		int iFrom = -1;
		int i     = 0;
		for(Iterator<Link> iter = path.iterator(); iter.hasNext(); i++ ) {
			Link e = iter.next();
			if (e.getSource().equals(from)) {
				iFrom = i;
			}
			if (e.getDest().equals(to)) {
				if(iFrom >= 0) {
					return path.subList(iFrom, i+1);
				} else {
					return Collections.emptyList();
				}
			}
		}
		return Collections.emptyList();
	}

	/**
	 * Returns a string representation of the flow.
	 * @return a string representation of the flow
	 */
	@Override
	public String toString() {
		if ( path.size() > 1 ) {
			return "(" + id + "," + source
			        + "," + sink + "," + ac
			        + "," + path + ")";			
		} else {
			return "(" + id + "," + path.get( 0 ).getSource().toString() + ac + ")";
		}
		
	}
}