/*
 *  (c) 2011 Steffen Bondorf, disco | Distributed Computer Systems Lab
 *                                  University of Kaiserslautern, Germany
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.network;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import unikl.disco.misc.SetUtils;
import unikl.disco.nc.Configuration;
import unikl.disco.curves.Curve;

/**
 * 
 * @author Steffen Bondorf
 *
 */

//TODO throw more meaningful exceptions

public class Network {
	private Set<Server> servers;
	private Set<Link> links;
	private Set<Flow> flows;

	private Server explicit_sink;
	
	private Map<Server,Set<Link>> map__server__in_links;
	private Map<Server,Set<Link>> map__server__out_links;

	private Map<Link,Set<Flow>> map__link__flows;
	private Map<Server,Set<Flow>> map__server__source_flows;
	
	public Network() {
		servers = new HashSet<Server>();
		links = new HashSet<Link>();
		flows = new HashSet<Flow>();
		
		map__server__in_links = new HashMap<Server,Set<Link>>();
		map__server__out_links = new HashMap<Server,Set<Link>>();

		map__server__source_flows = new HashMap<Server,Set<Flow>>();
		map__link__flows = new HashMap<Link,Set<Flow>>();

		explicit_sink = addDummyServer();
	}

	// ---------------------------------------------------------------------
	// servers
	// ---------------------------------------------------------------------
	public Server addServer( Curve service_curve ) {
		Server s = new Server( service_curve.copy() );
		
		map__server__in_links.put( s, new HashSet<Link>() );
		map__server__out_links.put( s, new HashSet<Link>() );

		servers.add( s );
		
		return s;
	}
	
	public Server addServer( Curve service_curve, Curve max_service_curve ) {
		Server s = new Server( service_curve.copy(), max_service_curve.copy() );
		
		map__server__in_links.put( s, new HashSet<Link>() );
		map__server__out_links.put( s, new HashSet<Link>() );

		servers.add( s );
		
		return s;
	}
	
	private Server addDummyServer() {
		Server s = new Server();
		
		map__server__in_links.put( s, new HashSet<Link>() );
		map__server__out_links.put( s, new HashSet<Link>() );

		servers.add( s );
		
		return s;
	}

	public boolean removeServer( Server s ) {
		boolean result = true;
		
		for( Link l : getIncidentLinks( s ) ) {
			for (Flow f: map__link__flows.get( l ) ){
				removeFlow( f );
			}
			result = result & removeLink( l );
		}
		map__server__in_links.remove( s );
		map__server__out_links.remove( s );

		// All flows should already be deleted
		for ( Flow f : map__server__source_flows.get( s ) ){
			result = result & removeFlow( f );
		}
		result = result & servers.remove( s );
		map__server__source_flows.remove( s );
		
		return result;
	}
	
	protected void addSourceFlow( Server source, Flow flow ){
	   Set<Flow> sfs = map__server__source_flows.get( source );
	   if ( sfs == null ) {
	       sfs = new HashSet<Flow>();
	       map__server__source_flows.put( source, sfs );
	   }
	   sfs.add( flow );
	}
	
	public Set<Flow> getSourceFlowSet( Server source ) {
		Set<Flow> sfs = map__server__source_flows.get( source );
		if ( sfs == null ){
			return new HashSet<Flow>();
		} else {
			return sfs;
		}
	}
	
	public Set<Server> getServers() {
		return servers;
	}
	
	public int numServers() {
		return servers.size();
	}

	public int degree( Server s ) {
		return inDegree( s ) + outDegree( s );
	}

	public int inDegree( Server s ) {
		return getInLinks( s ).size();
	}

	public int outDegree( Server s ) {
		return getOutLinks( s ).size();
	}

	public Set<Link> getInLinks( Server s ) {
		return map__server__in_links.get( s );
	}

	public Set<Link> getOutLinks( Server s ) {
		return map__server__out_links.get( s );
	}

	public Set<Link> getIncidentLinks( Server s ) {
		return SetUtils.getUnion( getInLinks( s ), getOutLinks( s ) );
	}

	public Set<Server> getNeighbors( Server s ) {
		return SetUtils.getUnion( getSuccessors( s ), getPredecessors( s ) );
	}

	public Set<Server> getSuccessors( Server s ) {
		Set<Server> successors = new HashSet<Server>();
		for( Link l : getOutLinks( s ) ) {
			successors.add( l.getDest() );
		}
		return successors;
	}

	public Set<Server> getPredecessors( Server s ) {
		Set<Server> predecessors = new HashSet<Server>();
		for( Link l : getInLinks( s ) ) {
			predecessors.add( l.getSource() );
		}
		return predecessors;
	}

	// -----------------------------------------------------------------------------------------
	// links
	// -----------------------------------------------------------------------------------------
	public Set<Link> getLinks() {
		return links;
	}
	
	// TODO make somehow explicit that a link from source to destination was already in the network
	//		and thus this link was returned instead of a newly created one
	public Link addLink( Server source, Server destination ) throws Exception {
		if( !servers.contains( source ) || !servers.contains( destination ) ){
			throw new Exception();
		}
		
		Link existing_link = findLink( source, destination );
		if( existing_link == null ) {
			
			Link new_link = new Link( source, destination );
			
			map__link__flows.put( new_link, new HashSet<Flow>() );
			
			map__server__in_links.get( new_link.getDest() ).add( new_link );
			map__server__out_links.get( new_link.getSource() ).add( new_link );
			
			links.add( new_link );
			return new_link;
		}else{
			return existing_link;
		}
	}

	public boolean removeLink( Link l ) {
		if( l == null ){ return false; }
		
		boolean result = true;
		for( Flow f : map__link__flows.get( l ) ) {
			result = result & removeFlow( f );
		}

		// remove does not return an boolean value
		map__link__flows.remove( l );
		
		for( Set<Link> links : map__server__in_links.values() ) {
			links.remove( l );
		}
		for( Set<Link> links : map__server__out_links.values() ) {
			links.remove( l );
		}
		
		result = result & links.remove( l );
		
		return result;
	}

	public Link findLink( Server src, Server dest ) {
		Set<Link> set_connecting_links = SetUtils.getIntersection( getInLinks( dest ), getOutLinks( src ) );
		if( set_connecting_links.isEmpty() ) {
			return null;
		}else{
			if( set_connecting_links.size() > 1 ) {
				return null;
			}else{
				return set_connecting_links.iterator().next();
			}
		}
	}
	
	/**
	 * Returns a set containing all flows traversing the given link
	 * <code>e</code>.
	 * 
	 * @param l the traversed link
	 * @return a set of traversing flows
	 */
	public Set<Flow> getLinkFlowSet( Link l ) {
		return map__link__flows.get( l );
	}
	
	// -----------------------------------------------------------------------------------------
	// flows
	// It is not allowed to change the path of a flow,
	// i.e., the user needs to remove the old flow and insert a new one with the modified path
	// -----------------------------------------------------------------------------------------
	public Set<Flow> getFlows() {
		return flows;
	}
	
	public Flow addFlow( Curve arrival_curve, List<Link> path ) throws Exception {
		try {
			return addFlow( arrival_curve, new LinkedList<Link>( path ), true );
		} catch (Exception e) {
			throw e;
		}
	}
	
	/**
	 * Adds a flow to the network. The flow instance has to be fully
	 * initialized, including the flow's path.
	 * 
	 * @param flow the flow to be added
	 */
	public Flow addFlow( Curve arrival_curve, Server source, Server sink ) throws Exception {
		List<Link> path = this.getShortestPath( source, sink );
		
		if( source == sink ) {
			return addFlow( arrival_curve, sink );
		}
		
		try {
			return addFlow( arrival_curve, path, false );
		} catch (Exception e) {
			throw e;
		}
	}
	
	private Flow addFlow( Curve arrival_curve, List<Link> path, boolean check_path ) throws Exception {
		// TODO check order of links and reorder of necessary/possible
		// TODO check given path for possible cyclic dependencies?
		
		if( Configuration.useCheckFlowPath() && !links.containsAll( path ) ) {
			throw new Exception();
		} else {
			Link sink_link = addLink( path.get( path.size() -1 ).getDest(), explicit_sink );

			path.add( sink_link );
			Flow f = new Flow( arrival_curve.copy(), path );

			flows.add( f );
			addSourceFlow( path.get( 0 ).getSource(), f );
			
			for ( Link l : path )
			{
				map__link__flows.get( l ).add( f );
			}
			
			return f;	
		}
	}
	
	public Flow addFlow( Curve arrival_curve, Server single_hop ) throws Exception {
		Link sink_link = addLink( single_hop, explicit_sink );
		
		List<Link> path = new LinkedList<Link>();
		path.add( sink_link );
		Flow f = new Flow( arrival_curve.copy(), path );
		
		flows.add( f );
		addSourceFlow( single_hop, f );
		
		for ( Iterator<Link> iter = path.iterator(); iter.hasNext(); )
		{
			map__link__flows.get( iter.next() ).add( f );
		}		
		
		return f;
	}
	
	/**
	 * Removes a flow from the network.
	 * 
	 * @param flow the flow to be removed
	 * @result false if something went wrong and the network may be in an inconsistent state
	 */
	public boolean removeFlow( Flow flow )
	{
		boolean result = true;
		for ( Link l : flow.getPathInclExplSnk() )
		{
			// single &: both sides will always be evaluated
			result = result & map__link__flows.get( l ).remove( flow );
		}

		Set<Flow> sflows = map__server__source_flows.get( flow.getSource() );
		result = result & sflows.remove( flow );
		
		result = result & flows.remove( flow );
		return result;
	}

	/**
	 * Returns a set containing all flows traversing the link from
	 * <code>predecessor</code> to <code>server</code>.
	 * 
	 * @param predecessor
	 * @param server
	 * @return a set of incoming flows
	 */
	public Set<Flow> getIncomingFlows( Server predecessor, Server server )
	{
		return map__link__flows.get( findLink( predecessor, server ) );
	}

	/**
	 * Returns the set of all flows that enter <code>server</code> from one of
	 * its predecessors.
	 * 
	 * @param server
	 * @return a set of all incoming flows
	 */
	public Set<Flow> getAllIncomingFlows( Server server )
	{
		HashSet<Flow> all_flows = new HashSet<Flow> (getSourceFlowSet( server ) );
		for ( Link l : getInLinks( server ) )
		{
			all_flows.addAll( map__link__flows.get( l ) );
		}
		return all_flows;
	}

	/**
	 * Returns an aggregate arrival curve for all flows originating in
	 * <code>source</code> and contained in the set <code>outgoing_flows</code>.
	 * 
	 * Returns a null curve if the intersection of those sets of curves is empty.
	 * 
	 * @param source
	 *            the source of all flows to be aggregated
	 * @param outgoing_flows
	 *            the set of all flows to be aggregated
	 * @return an aggregate arrival curve
	 */
	public Curve getAggregateSourceFlowAC( Server source, Set<Flow> outgoing_flows )
	{
		Curve a_out = Curve.createNullCurve();
		for ( Flow f : SetUtils.getIntersection( map__server__source_flows.get( source ), outgoing_flows ) )
		{
			a_out = Curve.add( a_out, f.getArrivalCurve() );
		}
		return a_out;
	}

	/**
	 * Returns the predecessor server from which a given <code>server</code>
	 * receives the flow <code>f</code>.
	 * 
	 * @param server
	 *            the receiving server
	 * @param f
	 *            the incoming flow
	 * @return the predecessor
	 */
	public Server getPredecessor( Server server, Flow f )
	{
		for ( Link l : getInLinks(server) )
		{
			if ( map__link__flows.get(l).contains( f ) )
			{
				return l.getSource();
			}
		}
		return null;
	}
 	
 	/**
 	 * Aggregates a set of flows by removing them from the internal management
 	 * structures and replacing it with a single dummy aggregate flow.
 	 * 
 	 * @param aggregateable_flows
 	 *            the flows to aggregate
 	 * @param joining_flows
 	 *            the set of flows with the same ingress as the aggregate
 	 * @param all_interfering_flows
 	 *            the set of all interfering flows
 	 * @param flow_egress_map
 	 *            a map of flows to egress servers
 	 * @return the aggregate flow.
 	 */
 	public Flow aggregateFlows( Set<Flow> aggregateable_flows,
 								Set<Flow> joining_flows,
 								Set<Flow> all_interfering_flows,
 								Map<Flow, Integer> flow_egress_map )
 	{
 		Flow aggregate_flow;

 		// backup egress server of the aggregateable flows
 		Integer egress_server = flow_egress_map.get( aggregateable_flows.iterator().next() );

 		// remove aggregateable flows from all local data structures...
 		joining_flows.removeAll( aggregateable_flows );
 		all_interfering_flows.removeAll( aggregateable_flows );
 		for ( Flow f : aggregateable_flows )
 		{
 			flow_egress_map.remove( f );
 		}

 		// ...and create a new aggregate flow instead.
 		aggregate_flow = Flow.createDummyFlow();
 		joining_flows.add( aggregate_flow );
 		all_interfering_flows.add( aggregate_flow );
 		flow_egress_map.put( aggregate_flow, egress_server );

 		return aggregate_flow;
 	}

 	/**
 	 * Creates a list containing the joining flows sets for each server along
 	 * the path. Does not include the flows_of_interest.
 	 * 
 	 * @param path
 	 * @param flows_of_interest
 	 * @return a list containing sets of joining flows
 	 */
 	public List<Set<Flow>> getJoiningFlowLists(List<Link> path, Set<Flow> flows_of_interest )
 	{
 		Set<Flow> present_flows = new HashSet<Flow>( flows_of_interest );
		List<Set<Flow>> joining_flow_sets = new ArrayList<Set<Flow>>();
 		for ( Link l : path )
 		{
 			Server server = l.getSource();

 			Set<Flow> joining_flows = SetUtils.getDifference( getAllIncomingFlows( server ), present_flows );
 			joining_flow_sets.add( joining_flows );

 			present_flows.addAll( joining_flows );

 			Set<Flow> leaving_flows = new HashSet<Flow>();
 			for ( Link l_out : getOutLinks(server) )
 			{
 				if ( l_out == l )
 				{
 					// skip outgoing link which is on path
 					continue;
 				}
 				leaving_flows.addAll( map__link__flows.get( l_out ) );
 			}
 			present_flows = SetUtils.getDifference( present_flows, leaving_flows );
 		}
 		return joining_flow_sets;
 	}

 	/**
 	 * Creates a list containing a set of rejoining flows for each server along
 	 * the path. Rejoining flows are flows that have joined the flow of interest
 	 * before, then took a different path and finally join the flow of interest
 	 * again.
 	 * 
 	 * @param joining_flow_sets
 	 *            the list of joining flows
 	 * @return a list containing sets of joining flows
 	 */
 	public List<Set<Flow>> getRejoiningFlowLists( List<Set<Flow>> joining_flow_sets )
 	{
 		Set<Flow> previously_joined_flows = new HashSet<Flow>();
 		List<Set<Flow>> rejoining_flow_lists = new LinkedList<Set<Flow>>();
 		for ( int i = 0; i < joining_flow_sets.size(); i++ )
 		{
 			Set<Flow> rejoining_flows = SetUtils.getIntersection( previously_joined_flows, joining_flow_sets.get( i ) );
 			rejoining_flow_lists.add( rejoining_flows );
 			previously_joined_flows.addAll( joining_flow_sets.get( i ) );
 		}
 		return rejoining_flow_lists;
 	}

 	/**
 	 * Computes the map of flows to egress point (an Integer index into the list
 	 * of servers on the path).
 	 * 
 	 * @param path
 	 * @param all_cross_traffic_flows
 	 * @return the map
 	 */
 	public Map<Flow, Integer> getFlowEgressMap(	List<Link> path, Set<Flow> all_cross_traffic_flows )
 	{
 		Map<Flow, Integer> map = new HashMap<Flow, Integer>();
 		int i = -1;
 		for ( Link l : path )
 		{
 			i++;
 			Server server = l.getSource();
 			Set<Flow> outgoing_flows = new HashSet<Flow>();
 			for ( Link l_out : getOutLinks( server ) )
 			{
 				if ( l_out == l )
 				{
 					// skip outgoing link which is on path
 					continue;
 				}
 				outgoing_flows.addAll( map__link__flows.get(l_out) );
 			}
 			outgoing_flows = SetUtils.getIntersection( outgoing_flows, all_cross_traffic_flows );
 			for ( Flow f : outgoing_flows )
 			{
 				if ( !map.containsKey( f ) )
 				{
 					map.put( f, Integer.valueOf( i ) ); // store only the first
 														// egress point of a flow
 				}
 			}
 		}
 		for ( Flow f : all_cross_traffic_flows )
 		{
 			if ( !map.containsKey( f ) )
 			{
 				map.put( f, Integer.valueOf( path.size() - 1 ) );
 			}
 		}
 		return map;
 	}

 	/**
 	 * Creates a partition of the set of flows joining this server so that each
 	 * partition contains flows with identical egress point.
 	 * 
 	 * @param path
 	 * @param server
 	 *            the ingress server
 	 * @param joining_flows
 	 *            the set of flows joining at the ingress server
 	 * @param flow_egress_map
 	 *            the flow-to-egress-point map
 	 * @return a set of flow sets with same egress point
 	 */
 	public Set<Set<Flow>> getAggregateableFlowSets(	List<Link> path,
 														int server,
 														Set<Flow> joining_flows,
 														Map<Flow, Integer> flow_egress_map )
 	{
 		Set<Set<Flow>> aggregateable_flows = new HashSet<Set<Flow>>();
 		List<Flow> remaining_flows = new LinkedList<Flow>( joining_flows );

 		while ( remaining_flows.size() > 0 )
 		{
 			Flow flow = remaining_flows.remove( remaining_flows.size() - 1 );

 			Integer common_egress_server = flow_egress_map.get( flow );
 			if ( common_egress_server == null || common_egress_server.intValue() < server )
 			{ // i.e. this flow is rejoining
 				common_egress_server = Integer.valueOf( findEgressPoint( path, server, flow ) );
 			}

 			Set<Flow> flow_set = new HashSet<Flow>( Collections.singleton( flow ) );
 			for ( int i = remaining_flows.size() - 1; i >= 0; i-- )
 			{
 				flow = remaining_flows.get( i );

 				Integer egress_server = flow_egress_map.get( flow );
 				if ( egress_server == null || egress_server.intValue() < server )
 				{ // i.e. this flow is rejoining
 					egress_server = Integer.valueOf( findEgressPoint( path, server, flow ) );
 				}

 				if ( egress_server.intValue() == common_egress_server.intValue() )
 				{
 					flow_set.add( flow );
 					remaining_flows.remove( i );
 				}
 			}
 			aggregateable_flows.add( flow_set );
 		}

 		return aggregateable_flows;
 	}
 	
	/**
	 * 
	 * @param src
	 * @param snk
	 * @return
	 */
    public List<Link> getShortestPath( Server src, Server snk ) {
		Set<Server> visited = new HashSet<Server>();
		Map<Server,List<Link>> paths = new HashMap<Server,List<Link>>();
		
		paths.put( src, new LinkedList<Link>() );
		
		LinkedList<Server> queue = new LinkedList<Server>();
		queue.add(src);
		visited.add(src);
		
		while ( !queue.isEmpty() ) {
			Server s = queue.getLast();
			queue.remove( s );

			for( Server successor : getSuccessors( s ) ) {

				List<Link> path_tmp = new LinkedList<Link>( paths.get(s) );
				path_tmp.add( findLink( s, successor ) );
				
				if( !visited.contains(successor) ) {
					// add path
					paths.put( successor, path_tmp );
					queue.add( successor );
					visited.add( successor );
				}else{
					// compare path lengths
					if( paths.get( successor ).size() > path_tmp.size() ) {
						paths.put( successor, path_tmp );
						// TODO is it necessary to recheck the path to u?
						queue.add( successor );
					}
				}
			}
		}
		return paths.get( snk );
	}

 	/**
 	 * Returns the index of the server at which the given flow exits the path.
 	 * This point must be after the server specified by server_idx.
 	 * 
 	 * @param path
 	 *            the path
 	 * @param server_idx
 	 *            the index of the server where the flow joins the path
 	 * @param flow
 	 *            the flow
 	 * @return the index of the server in the path
 	 */
 	public int findEgressPoint( List<Link> path, int server_idx, Flow flow )
 	{
 		int i = -1;
 		for ( Link l : path )
 		{
 			i++;

 			if ( i < server_idx )
 			{
 				continue;
 			}

 			Server server = l.getSource();
 			for ( Link l_out : getOutLinks(server) )
 			{
 				if ( l_out == l )
 				{
 					// skip outgoing link which is on path
 					continue;
 				}
 				if ( map__link__flows.get( l_out ).contains( flow ) )
 				{
 					return i;
 				}
 			}
 		}
 		return path.size() - 1;
 	}

 	/**
 	 * Creates a fresh flow substitute for a rejoining flow and updates the
 	 * respective management information.
 	 * 
 	 * @param rejoining_flows_in_aggregate
 	 * @param egress_server
 	 * @param aggregateable_flows
 	 * @param all_interfering_flows
 	 * @param joining_flows
 	 * @param flow_egress_map
 	 */
 	public void createSubstituteFlows(	Set<Flow> rejoining_flows_in_aggregate,
 											int egress_server,
 											Set<Flow> aggregateable_flows,
 											Set<Flow> all_interfering_flows,
 											Set<Flow> joining_flows,
 											Map<Flow, Integer> flow_egress_map )
 	{
 		for ( Flow flow : rejoining_flows_in_aggregate )
 		{
 			Flow substitute = Flow.createDummyFlow();
 			aggregateable_flows.remove( flow );
 			aggregateable_flows.add( substitute );
 			all_interfering_flows.add( substitute );
 			joining_flows.remove( flow );
 			joining_flows.add( substitute );
 			flow_egress_map.put( substitute, Integer.valueOf( egress_server ) );
 		}
 	}

 	/**
 	 * Returns the server at which the flows in <code>flows_of_interest</code>
 	 * first all meet each other (when viewed from the source). When viewed from
 	 * <code>server</code> towards the sink, this is the last server where all
 	 * flows are still together.
 	 * 
 	 * @param server
 	 * @param flows_of_interest
 	 * @return the splitting point server
 	 */
 	public Server findSplittingPoint( Server server, Set<Flow> flows_of_interest )
 	{
 		Flow f = flows_of_interest.iterator().next();

 		List<Link> path = f.getSubPath( f.getSource(), server );

 		for ( Link l : path )
 		{
 			Set<Flow> flow_list = map__link__flows.get(l);
 			if ( SetUtils.getIntersection( flow_list, flows_of_interest ).size() == flows_of_interest.size() )
 			{
 				return l.getSource();
 			}
 		}
 		return server;
 	}
}
